import { __rest, __assign } from 'tslib';
import { invariant } from 'hey-listen';
import { PanSession } from '../PanSession.js';
import { getGlobalLock } from './utils/lock.js';
import { isRefObject } from '../../utils/is-ref-object.js';
import { addPointerEvent } from '../../events/use-pointer-event.js';
import { addDomEvent } from '../../events/use-dom-event.js';
import { calcRelativeConstraints, resolveDragElastic, calcViewportConstraints, applyConstraints, rebaseAxisConstraints, defaultElastic } from './utils/constraints.js';
import { AnimationType } from '../../render/utils/types.js';
import { createBox } from '../../projection/geometry/models.js';
import { eachAxis } from '../../projection/utils/each-axis.js';
import { measureViewportBox } from '../../projection/node/HTMLProjectionNode.js';
import { convertBoxToBoundingBox, convertBoundingBoxToBox } from '../../projection/geometry/conversion.js';

var elementDragControls = new WeakMap();
/**
 *
 */
var lastPointerEvent;
var VisualElementDragControls = /** @class */ (function () {
    function VisualElementDragControls(_a) {
        var visualElement = _a.visualElement;
        /**
         * Track whether we're currently dragging.
         *
         * @internal
         */
        this.isDragging = false;
        /**
         * The current direction of drag, or `null` if both.
         *
         * @internal
         */
        this.currentDirection = null;
        /**
         * The permitted boundaries of travel, in pixels.
         *
         * @internal
         */
        this.constraints = false;
        /**
         * The per-axis resolved elastic values.
         *
         * @internal
         */
        this.elastic = createBox();
        /**
         * A reference to the host component's latest props.
         *
         * @internal
         */
        this.props = {};
        /**
         * @internal
         */
        this.hasMutatedConstraints = false;
        /**
         * Track the initial position of the cursor relative to the dragging element
         * when dragging starts as a value of 0-1 on each axis. We then use this to calculate
         * an ideal bounding box for the VisualElement renderer to project into every frame.
         *
         * @internal
         */
        this.cursorProgress = {
            x: 0.5,
            y: 0.5,
        };
        // When updating _dragX, or _dragY instead of the VisualElement,
        // persist their values between drag gestures.
        this.originPoint = {};
        // This is a reference to the global drag gesture lock, ensuring only one component
        // can "capture" the drag of one or both axes.
        // TODO: Look into moving this into pansession?
        this.openGlobalLock = null;
        /**
         * @internal
         */
        this.panSession = null;
        this.visualElement = visualElement;
        // this.visualElement.enableLayoutProjection()
        elementDragControls.set(visualElement, this);
    }
    /**
     * Instantiate a PanSession for the drag gesture
     *
     * @public
     */
    VisualElementDragControls.prototype.start = function (originEvent, _options) {
        var _this = this;
        var onSessionStart = function (_event) {
            // Stop any animations on both axis values immediately. This allows the user to throw and catch
            // the component.
            _this.stopMotion();
            /**
             * Save the initial point. We'll use this to calculate the pointer's position rather
             * than the one we receive when the gesture actually starts. By then, the pointer will
             * have already moved, and the perception will be of the pointer "slipping" across the element
             */
            // const initialPoint = getViewportPointFromEvent(event).point
            // this.cancelLayout?.()
            // this.cancelLayout = batchLayout((read, write) => {
            //     const ancestors = collectProjectingAncestors(this.visualElement)
            //     const children = collectProjectingChildren(this.visualElement)
            //     const tree = [...ancestors, ...children]
            //     let hasManuallySetCursorOrigin: boolean = false
            //     /**
            //      * Apply a simple lock to the projection target. This ensures no animations
            //      * can run on the projection box while this lock is active.
            //      */
            //     this.isLayoutDrag() && this.visualElement.lockProjectionTarget()
            //     write(() => {
            //         tree.forEach((element) => element.resetTransform())
            //     })
            //     read(() => {
            //         updateLayoutMeasurement(this.visualElement)
            //         children.forEach(updateLayoutMeasurement)
            //     })
            //     write(() => {
            //         tree.forEach((element) => element.restoreTransform())
            //         if (snapToCursor) {
            //             hasManuallySetCursorOrigin = this.snapToCursor(
            //                 initialPoint
            //             )
            //         }
            //     })
            //     read(() => {
            //         const isRelativeDrag = Boolean(
            //             this.getAxisMotionValue("x") && !this.isExternalDrag()
            //         )
            //         if (!isRelativeDrag) {
            //             this.visualElement.rebaseProjectionTarget(
            //                 true,
            //                 this.visualElement.measureViewportBox(false)
            //             )
            //         }
            //         this.visualElement.scheduleUpdateLayoutProjection()
            //         /**
            //          * When dragging starts, we want to find where the cursor is relative to the bounding box
            //          * of the element. Every frame, we calculate a new bounding box using this relative position
            //          * and let the visualElement renderer figure out how to reproject the element into this bounding
            //          * box.
            //          *
            //          * By doing it this way, rather than applying an x/y transform directly to the element,
            //          * we can ensure the component always visually sticks to the cursor as we'd expect, even
            //          * if the DOM element itself changes layout as a result of React updates the user might
            //          * make based on the drag position.
            //          */
            //         const { projection } = this.visualElement
            //         eachAxis((axis) => {
            //             if (!hasManuallySetCursorOrigin) {
            //                 const { min, max } = projection.target[axis]
            //                 this.cursorProgress[axis] = cursorProgress
            //                     ? cursorProgress[axis]
            //                     : progress(min, max, initialPoint[axis])
            //             }
            //             /**
            //              * If we have external drag MotionValues, record their origin point. On pointermove
            //              * we'll apply the pan gesture offset directly to this value.
            //              */
            //             const axisValue = this.getAxisMotionValue(axis)
            //             if (axisValue) {
            //                 this.originPoint[axis] = axisValue.get()
            //             }
            //         })
            //     })
            //     write(() => {
            //         flushSync.update()
            //         flushSync.preRender()
            //         flushSync.render()
            //         flushSync.postRender()
            //     })
            //     read(() => this.resolveDragConstraints())
            // })
        };
        var onStart = function (event, info) {
            var _a, _b, _c;
            // Attempt to grab the global drag gesture lock - maybe make this part of PanSession
            var _d = _this.props, drag = _d.drag, dragPropagation = _d.dragPropagation;
            if (drag && !dragPropagation) {
                if (_this.openGlobalLock)
                    _this.openGlobalLock();
                _this.openGlobalLock = getGlobalLock(drag);
                // If we don 't have the lock, don't start dragging
                if (!_this.openGlobalLock)
                    return;
            }
            // flushLayout()
            // Set current drag status
            _this.isDragging = true;
            _this.currentDirection = null;
            // Fire onDragStart event
            (_b = (_a = _this.props).onDragStart) === null || _b === void 0 ? void 0 : _b.call(_a, event, info);
            (_c = _this.visualElement.animationState) === null || _c === void 0 ? void 0 : _c.setActive(AnimationType.Drag, true);
        };
        var onMove = function (event, info) {
            var _a, _b, _c, _d;
            var _e = _this.props, dragPropagation = _e.dragPropagation, dragDirectionLock = _e.dragDirectionLock;
            // If we didn't successfully receive the gesture lock, early return.
            if (!dragPropagation && !_this.openGlobalLock)
                return;
            var offset = info.offset;
            // Attempt to detect drag direction if directionLock is true
            if (dragDirectionLock && _this.currentDirection === null) {
                _this.currentDirection = getCurrentDirection(offset);
                // If we've successfully set a direction, notify listener
                if (_this.currentDirection !== null) {
                    (_b = (_a = _this.props).onDirectionLock) === null || _b === void 0 ? void 0 : _b.call(_a, _this.currentDirection);
                }
                return;
            }
            // Update each point with the latest position
            _this.updateAxis("x", info.point, offset);
            _this.updateAxis("y", info.point, offset);
            // Fire onDrag event
            (_d = (_c = _this.props).onDrag) === null || _d === void 0 ? void 0 : _d.call(_c, event, info);
            // Update the last pointer event
            lastPointerEvent = event;
        };
        var onSessionEnd = function (event, info) {
            return _this.stop(event, info);
        };
        var transformPagePoint = this.props.transformPagePoint;
        this.panSession = new PanSession(originEvent, {
            onSessionStart: onSessionStart,
            onStart: onStart,
            onMove: onMove,
            onSessionEnd: onSessionEnd,
        }, { transformPagePoint: transformPagePoint });
    };
    VisualElementDragControls.prototype.resolveDragConstraints = function () {
        var _this = this;
        var _a = this.props, dragConstraints = _a.dragConstraints, dragElastic = _a.dragElastic;
        // const { layoutCorrected: layout } = this.visualElement.getLayoutState()
        var layout = createBox();
        if (dragConstraints) {
            this.constraints = isRefObject(dragConstraints)
                ? this.resolveRefConstraints(layout, dragConstraints)
                : calcRelativeConstraints(layout, dragConstraints);
        }
        else {
            this.constraints = false;
        }
        this.elastic = resolveDragElastic(dragElastic);
        /**
         * If we're outputting to external MotionValues, we want to rebase the measured constraints
         * from viewport-relative to component-relative.
         */
        if (this.constraints && !this.hasMutatedConstraints) {
            eachAxis(function (axis) {
                if (_this.getAxisMotionValue(axis)) {
                    _this.constraints[axis] = rebaseAxisConstraints(layout[axis], _this.constraints[axis]);
                }
            });
        }
    };
    VisualElementDragControls.prototype.resolveRefConstraints = function (layoutBox, constraints) {
        var _a = this.props, onMeasureDragConstraints = _a.onMeasureDragConstraints, transformPagePoint = _a.transformPagePoint;
        var constraintsElement = constraints.current;
        invariant(constraintsElement !== null, "If `dragConstraints` is set as a React ref, that ref must be passed to another component's `ref` prop.");
        this.constraintsBox = measureViewportBox(constraintsElement, transformPagePoint);
        var measuredConstraints = calcViewportConstraints(layoutBox, this.constraintsBox);
        /**
         * If there's an onMeasureDragConstraints listener we call it and
         * if different constraints are returned, set constraints to that
         */
        if (onMeasureDragConstraints) {
            var userConstraints = onMeasureDragConstraints(convertBoxToBoundingBox(measuredConstraints));
            this.hasMutatedConstraints = !!userConstraints;
            if (userConstraints) {
                measuredConstraints = convertBoundingBoxToBox(userConstraints);
            }
        }
        return measuredConstraints;
    };
    VisualElementDragControls.prototype.cancelDrag = function () {
        var _a, _b;
        // this.visualElement.unlockProjectionTarget()
        (_a = this.cancelLayout) === null || _a === void 0 ? void 0 : _a.call(this);
        this.isDragging = false;
        this.panSession && this.panSession.end();
        this.panSession = null;
        if (!this.props.dragPropagation && this.openGlobalLock) {
            this.openGlobalLock();
            this.openGlobalLock = null;
        }
        (_b = this.visualElement.animationState) === null || _b === void 0 ? void 0 : _b.setActive(AnimationType.Drag, false);
    };
    VisualElementDragControls.prototype.stop = function (event, info) {
        var _a, _b, _c;
        (_a = this.panSession) === null || _a === void 0 ? void 0 : _a.end();
        this.panSession = null;
        var isDragging = this.isDragging;
        this.cancelDrag();
        if (!isDragging)
            return;
        var velocity = info.velocity;
        this.animateDragEnd(velocity);
        (_c = (_b = this.props).onDragEnd) === null || _c === void 0 ? void 0 : _c.call(_b, event, info);
    };
    VisualElementDragControls.prototype.snapToCursor = function (point) {
        var _this = this;
        return eachAxis(function (axis) {
            var drag = _this.props.drag;
            // If we're not dragging this axis, do an early return.
            if (!shouldDrag(axis, drag, _this.currentDirection))
                return;
            var axisValue = _this.getAxisMotionValue(axis);
            if (axisValue) {
                // const box = this.visualElement.getLayoutState().layout
                var box = createBox();
                var length_1 = box[axis].max - box[axis].min;
                var center = box[axis].min + length_1 / 2;
                var offset = point[axis] - center;
                _this.originPoint[axis] = point[axis];
                axisValue.set(offset);
            }
            else {
                _this.cursorProgress[axis] = 0.5;
                return true;
            }
        }); //.includes(true)
    };
    /**
     * Update the specified axis with the latest pointer information.
     */
    VisualElementDragControls.prototype.updateAxis = function (axis, point, offset) {
        var drag = this.props.drag;
        // If we're not dragging this axis, do an early return.
        if (!shouldDrag(axis, drag, this.currentDirection))
            return;
        return this.getAxisMotionValue(axis)
            ? this.updateAxisMotionValue(axis, offset)
            : this.updateVisualElementAxis(axis, point);
    };
    VisualElementDragControls.prototype.updateAxisMotionValue = function (axis, offset) {
        var axisValue = this.getAxisMotionValue(axis);
        if (!offset || !axisValue)
            return;
        var nextValue = this.originPoint[axis] + offset[axis];
        var update = this.constraints
            ? applyConstraints(nextValue, this.constraints[axis], this.elastic[axis])
            : nextValue;
        axisValue.set(update);
    };
    VisualElementDragControls.prototype.updateVisualElementAxis = function (_axis, _point) {
        // Get the actual layout bounding box of the element
        // const axisLayout = createBox().x
        // const axisLayout = this.visualElement.getLayoutState().layout[axis]
        // // Calculate its current length. In the future we might want to lerp this to animate
        // // between lengths if the layout changes as we change the DOM
        // const axisLength = axisLayout.max - axisLayout.min
        // // Get the initial progress that the pointer sat on this axis on gesture start.
        // const axisProgress = this.cursorProgress[axis]
        // // Calculate a new min point based on the latest pointer position, constraints and elastic
        // const min = calcConstrainedMinPoint(
        //     point[axis],
        //     axisLength,
        //     axisProgress,
        //     this.constraints?.[axis],
        //     this.elastic[axis]
        // )
        // Update the axis viewport target with this new min and the length
        // this.visualElement.setProjectionTargetAxis(axis, min, min + axisLength)
    };
    VisualElementDragControls.prototype.setProps = function (_a) {
        var _b = _a.drag, drag = _b === void 0 ? false : _b, _c = _a.dragDirectionLock, dragDirectionLock = _c === void 0 ? false : _c, _d = _a.dragPropagation, dragPropagation = _d === void 0 ? false : _d, _e = _a.dragConstraints, dragConstraints = _e === void 0 ? false : _e, _f = _a.dragElastic, dragElastic = _f === void 0 ? defaultElastic : _f, _g = _a.dragMomentum, dragMomentum = _g === void 0 ? true : _g, remainingProps = __rest(_a, ["drag", "dragDirectionLock", "dragPropagation", "dragConstraints", "dragElastic", "dragMomentum"]);
        this.props = __assign({ drag: drag,
            dragDirectionLock: dragDirectionLock,
            dragPropagation: dragPropagation,
            dragConstraints: dragConstraints,
            dragElastic: dragElastic,
            dragMomentum: dragMomentum }, remainingProps);
    };
    /**
     * Drag works differently depending on which props are provided.
     *
     * - If _dragX and _dragY are provided, we output the gesture delta directly to those motion values.
     * - If the component will perform layout animations, we output the gesture to the component's
     *      visual bounding box
     * - Otherwise, we apply the delta to the x/y motion values.
     */
    VisualElementDragControls.prototype.getAxisMotionValue = function (axis) {
        var _a = this.props, layout = _a.layout, layoutId = _a.layoutId;
        var dragKey = "_drag" + axis.toUpperCase();
        if (this.props[dragKey]) {
            return this.props[dragKey];
        }
        else if (!layout && layoutId === undefined) {
            return this.visualElement.getValue(axis, 0);
        }
    };
    // private isLayoutDrag() {
    //     return !this.getAxisMotionValue("x")
    // }
    // private isExternalDrag() {
    //     const { _dragX, _dragY } = this.props
    //     return _dragX || _dragY
    // }
    VisualElementDragControls.prototype.animateDragEnd = function (_velocity) {
        // const { drag, dragMomentum, dragElastic, dragTransition } = this.props
        /**
         * Everything beyond the drag gesture should be performed with
         * relative projection so children stay in sync with their parent element.
         */
        // const isRelative = convertToRelativeProjection(
        //     this.visualElement,
        //     this.isLayoutDrag() && !this.isExternalDrag()
        // )
        // /**
        //  * If we had previously resolved constraints relative to the viewport,
        //  * we need to also convert those to a relative coordinate space for the animation
        //  */
        // const constraints = this.constraints || {}
        // if (
        //     isRelative &&
        //     Object.keys(constraints).length &&
        //     this.isLayoutDrag()
        // ) {
        //     const projectionParent = this.visualElement.getProjectionParent()
        //     if (projectionParent) {
        //         const relativeConstraints = calcRelativeOffset(
        //             projectionParent.projection.targetFinal,
        //             constraints as Box
        //         )
        //         eachAxis((axis) => {
        //             const { min, max } = relativeConstraints[axis]
        //             constraints[axis] = {
        //                 min: isNaN(min) ? undefined : min,
        //                 max: isNaN(max) ? undefined : max,
        //             }
        //         })
        //     }
        // }
        // const momentumAnimations = eachAxis((axis) => {
        //     if (!shouldDrag(axis, drag, this.currentDirection)) {
        //         return
        //     }
        //     const transition = constraints?.[axis] ?? {}
        //     /**
        //      * Overdamp the boundary spring if `dragElastic` is disabled. There's still a frame
        //      * of spring animations so we should look into adding a disable spring option to `inertia`.
        //      * We could do something here where we affect the `bounceStiffness` and `bounceDamping`
        //      * using the value of `dragElastic`.
        //      */
        //     const bounceStiffness = dragElastic ? 200 : 1000000
        //     const bounceDamping = dragElastic ? 40 : 10000000
        //     const inertia = {
        //         type: "inertia",
        //         velocity: dragMomentum ? velocity[axis] : 0,
        //         bounceStiffness,
        //         bounceDamping,
        //         timeConstant: 750,
        //         restDelta: 1,
        //         restSpeed: 10,
        //         ...dragTransition,
        //         ...transition,
        //     }
        //     // If we're not animating on an externally-provided `MotionValue` we can use the
        //     // component's animation controls which will handle interactions with whileHover (etc),
        //     // otherwise we just have to animate the `MotionValue` itself.
        //     return this.getAxisMotionValue(axis)
        //         ? this.startAxisValueAnimation(axis, inertia)
        //         : this.visualElement.startLayoutAnimation(
        //               axis,
        //               inertia,
        //               isRelative
        //           )
        //     return true
        // })
        return Promise.resolve();
        // Run all animations and then resolve the new drag constraints.
        // return Promise.all(momentumAnimations).then(() => {
        //     this.props.onDragTransitionEnd?.()
        // })
    };
    VisualElementDragControls.prototype.stopMotion = function () {
        // eachAxis((axis) => {
        // const axisValue = this.getAxisMotionValue(axis)
        // axisValue
        //     ? axisValue.stop()
        //     : this.visualElement.stopLayoutAnimation()
        // })
    };
    // private startAxisValueAnimation(axis: "x" | "y", transition: Transition) {
    //     const axisValue = this.getAxisMotionValue(axis)
    //     if (!axisValue) return
    //     const currentValue = axisValue.get()
    //     axisValue.set(currentValue)
    //     axisValue.set(currentValue) // Set twice to hard-reset velocity
    //     return startAnimation(axis, axisValue, 0, transition)
    // }
    VisualElementDragControls.prototype.scalePoint = function () {
        // const { drag, dragConstraints } = this.props
        // if (!isRefObject(dragConstraints) || !this.constraintsBox) return
        // // Stop any current animations as there can be some visual glitching if we resize mid animation
        // this.stopMotion()
        // // Record the relative progress of the targetBox relative to the constraintsBox
        // const boxProgress = { x: 0, y: 0 }
        // eachAxis((axis) => {
        //     boxProgress[axis] = calcOrigin(
        //         this.visualElement.projection.target[axis],
        //         this.constraintsBox![axis]
        //     )
        // })
        /**
         * For each axis, calculate the current progress of the layout axis within the constraints.
         * Then, using the latest layout and constraints measurements, reposition the new layout axis
         * proportionally within the constraints.
         */
        // this.updateConstraints(() => {
        //     eachAxis((axis) => {
        //         if (!shouldDrag(axis, drag, null)) return
        //         // Calculate the position of the targetBox relative to the constraintsBox using the
        //         // previously calculated progress
        //         const { min, max } = calcPositionFromProgress(
        //             this.visualElement.projection.target[axis],
        //             this.constraintsBox![axis],
        //             boxProgress[axis]
        //         )
        //         this.visualElement.setProjectionTargetAxis(axis, min, max)
        //     })
        // })
        /**
         * If any other draggable components are queuing the same tasks synchronously
         * this will wait until they've all been scheduled before flushing.
         */
        // setTimeout(flushLayout, 1)
    };
    VisualElementDragControls.prototype.updateConstraints = function (_onReady) {
        // this.cancelLayout = batchLayout((read, write) => {
        //     const ancestors = collectProjectingAncestors(this.visualElement)
        //     write(() =>
        //         ancestors.forEach((element) => element.resetTransform())
        //     )
        //     read(() => updateLayoutMeasurement(this.visualElement))
        //     write(() =>
        //         ancestors.forEach((element) => element.restoreTransform())
        //     )
        //     read(() => {
        //         this.resolveDragConstraints()
        //     })
        //     if (onReady) write(onReady)
        // })
    };
    VisualElementDragControls.prototype.mount = function (visualElement) {
        var _this = this;
        var element = visualElement.getInstance();
        /**
         * Attach a pointerdown event listener on this DOM element to initiate drag tracking.
         */
        var stopPointerListener = addPointerEvent(element, "pointerdown", function (event) {
            var _a = _this.props, drag = _a.drag, _b = _a.dragListener, dragListener = _b === void 0 ? true : _b;
            drag && dragListener && _this.start(event);
        });
        /**
         * Attach a window resize listener to scale the draggable target within its defined
         * constraints as the window resizes.
         */
        var stopResizeListener = addDomEvent(window, "resize", function () {
            _this.scalePoint();
        });
        /**
         * Ensure drag constraints are resolved correctly relative to the dragging element
         * whenever its layout changes.
         */
        var stopLayoutUpdateListener = visualElement.onLayoutUpdate(function () {
            if (_this.isDragging) {
                _this.resolveDragConstraints();
            }
        });
        /**
         * If the previous component with this same layoutId was dragging at the time
         * it was unmounted, we want to continue the same gesture on this component.
         */
        var prevDragCursor = visualElement.prevDragCursor;
        if (prevDragCursor) {
            this.start(lastPointerEvent, { cursorProgress: prevDragCursor });
        }
        /**
         * Return a function that will teardown the drag gesture
         */
        return function () {
            stopPointerListener === null || stopPointerListener === void 0 ? void 0 : stopPointerListener();
            stopResizeListener === null || stopResizeListener === void 0 ? void 0 : stopResizeListener();
            stopLayoutUpdateListener === null || stopLayoutUpdateListener === void 0 ? void 0 : stopLayoutUpdateListener();
            _this.cancelDrag();
        };
    };
    return VisualElementDragControls;
}());
function shouldDrag(direction, drag, currentDirection) {
    return ((drag === true || drag === direction) &&
        (currentDirection === null || currentDirection === direction));
}
/**
 * Based on an x/y offset determine the current drag direction. If both axis' offsets are lower
 * than the provided threshold, return `null`.
 *
 * @param offset - The x/y offset from origin.
 * @param lockThreshold - (Optional) - the minimum absolute offset before we can determine a drag direction.
 */
function getCurrentDirection(offset, lockThreshold) {
    if (lockThreshold === void 0) { lockThreshold = 10; }
    var direction = null;
    if (Math.abs(offset.y) > lockThreshold) {
        direction = "y";
    }
    else if (Math.abs(offset.x) > lockThreshold) {
        direction = "x";
    }
    return direction;
}

export { VisualElementDragControls, elementDragControls };
