import { addUniqueItem, removeItem } from '../../utils/array.js';

var NodeStack = /** @class */ (function () {
    function NodeStack() {
        this.members = [];
    }
    NodeStack.prototype.add = function (node) {
        // !this.members.size &&
        this.promote(node);
        addUniqueItem(this.members, node);
        node.scheduleRender();
    };
    NodeStack.prototype.remove = function (node) {
        removeItem(this.members, node);
        if (node === this.lead) {
            var prevLead = this.members[this.members.length - 1];
            if (prevLead) {
                this.promote(prevLead);
            }
        }
    };
    NodeStack.prototype.promote = function (node) {
        var prevLead = this.lead;
        this.lead = node;
        node.show();
        if (prevLead) {
            prevLead.scheduleRender();
            node.snapshot = prevLead.snapshot;
            if (node.snapshot) {
                node.snapshot.latestValues =
                    prevLead.animationValues || prevLead.latestValues;
            }
            node.isLayoutDirty = true;
            var crossfade = node.options.crossfade;
            if (crossfade === false) {
                prevLead.hide();
            }
            /**
             * TODO:
             *   - Test border radius when previous node was deleted
             *   - boxShadow mixing
             *   - Shared between element A in scrolled container and element B (scroll stays the same or changes)
             *   - Shared between element A in transformed container and element B (transform stays the same or changes)
             *   - Shared between element A in scrolled page and element B (scroll stays the same or changes)
             * ---
             *   - Crossfade opacity of root nodes
             *   - layoutId changes after animation
             *   - layoutId changes mid animation
             */
        }
    };
    return NodeStack;
}());

export { NodeStack };
