import { RefObject } from "react";
import { AnyPointerEvent, PanInfo } from "../../gestures/PanSession";
import { DraggableProps } from "./types";
import { VisualElement } from "../../render/types";
import { MotionProps } from "../../motion/types";
import { Box, Point, TransformPoint } from "../../projection/geometry/types";
export declare const elementDragControls: WeakMap<VisualElement<any, any>, VisualElementDragControls>;
interface DragControlConfig {
    visualElement: VisualElement;
}
export interface DragControlOptions {
    snapToCursor?: boolean;
    cursorProgress?: Point;
}
interface DragControlsProps extends DraggableProps {
    transformPagePoint?: TransformPoint;
}
declare type DragDirection = "x" | "y";
export declare class VisualElementDragControls {
    /**
     * Track whether we're currently dragging.
     *
     * @internal
     */
    isDragging: boolean;
    /**
     * The current direction of drag, or `null` if both.
     *
     * @internal
     */
    private currentDirection;
    /**
     * The permitted boundaries of travel, in pixels.
     *
     * @internal
     */
    private constraints;
    /**
     * The per-axis resolved elastic values.
     *
     * @internal
     */
    private elastic;
    /**
     * A reference to the host component's latest props.
     *
     * @internal
     */
    private props;
    /**
     * @internal
     */
    private visualElement;
    /**
     * @internal
     */
    private hasMutatedConstraints;
    /**
     * @internal
     */
    private cancelLayout?;
    /**
     * Track the initial position of the cursor relative to the dragging element
     * when dragging starts as a value of 0-1 on each axis. We then use this to calculate
     * an ideal bounding box for the VisualElement renderer to project into every frame.
     *
     * @internal
     */
    cursorProgress: Point;
    private originPoint;
    private openGlobalLock;
    /**
     * @internal
     */
    private panSession;
    /**
     * A reference to the measured constraints bounding box
     */
    private constraintsBox?;
    constructor({ visualElement }: DragControlConfig);
    /**
     * Instantiate a PanSession for the drag gesture
     *
     * @public
     */
    start(originEvent: AnyPointerEvent, _options?: DragControlOptions): void;
    resolveDragConstraints(): void;
    resolveRefConstraints(layoutBox: Box, constraints: RefObject<Element>): {
        x: {
            min: number;
            max: number;
        };
        y: {
            min: number;
            max: number;
        };
    };
    cancelDrag(): void;
    stop(event: AnyPointerEvent, info: PanInfo): void;
    snapToCursor(point: Point): void;
    /**
     * Update the specified axis with the latest pointer information.
     */
    updateAxis(axis: DragDirection, point: Point, offset?: Point): void;
    updateAxisMotionValue(axis: DragDirection, offset?: Point): void;
    updateVisualElementAxis(_axis: DragDirection, _point: Point): void;
    setProps({ drag, dragDirectionLock, dragPropagation, dragConstraints, dragElastic, dragMomentum, ...remainingProps }: DragControlsProps & MotionProps): void;
    /**
     * Drag works differently depending on which props are provided.
     *
     * - If _dragX and _dragY are provided, we output the gesture delta directly to those motion values.
     * - If the component will perform layout animations, we output the gesture to the component's
     *      visual bounding box
     * - Otherwise, we apply the delta to the x/y motion values.
     */
    private getAxisMotionValue;
    private animateDragEnd;
    stopMotion(): void;
    scalePoint(): void;
    updateConstraints(_onReady?: () => void): void;
    mount(visualElement: VisualElement): () => void;
}
export declare function expectsResolvedDragConstraints({ dragConstraints, onMeasureDragConstraints, }: MotionProps): boolean;
export {};
