import { AnimationOptions, AnimationPlaybackControls } from "../../animation/animate";
import { ResolvedValues } from "../../render/types";
import { SubscriptionManager } from "../../utils/subscription-manager";
import { AxisDelta, Box, Delta, Point } from "../geometry/types";
import { NodeStack } from "../shared/stack";
import { IProjectionNode, LayoutUpdateData, LayoutUpdateHandler, ProjectionNodeConfig, ProjectionNodeOptions, Snapshot } from "./types";
export declare function createProjectionNode<I>({ attachResizeListener, defaultParent, measureScroll, resetTransform, }: ProjectionNodeConfig<I>): {
    new (id: number, latestValues: ResolvedValues, parent?: IProjectionNode | undefined): {
        id: number;
        instance: I;
        root: IProjectionNode;
        parent?: IProjectionNode<unknown> | undefined;
        path: IProjectionNode[];
        children: Set<IProjectionNode<unknown>>;
        options: ProjectionNodeOptions;
        snapshot: Snapshot | undefined;
        layout: Box | undefined;
        layoutCorrected: Box;
        scroll?: Point | undefined;
        isLayoutDirty: boolean;
        isUpdating: boolean;
        shouldResetTransform: boolean;
        treeScale: Point;
        targetDelta?: Delta | undefined;
        projectionDelta?: Delta | undefined;
        projectionDeltaWithTransform?: Delta | undefined;
        lead?: IProjectionNode<unknown> | undefined;
        target?: Box | undefined;
        targetWithTransforms?: Box | undefined;
        latestValues: ResolvedValues;
        layoutWillUpdateListeners?: SubscriptionManager<VoidFunction> | undefined;
        layoutDidUpdateListeners?: SubscriptionManager<LayoutUpdateHandler> | undefined;
        potentialNodes: Map<number, IProjectionNode<unknown>>;
        registerPotentialNode(id: number, node: IProjectionNode): void;
        /**
         * Lifecycles
         */
        mount(instance: I, isLayoutDirty?: boolean): void;
        unmount(): void;
        startUpdate(): void;
        willUpdate(shouldNotifyListeners?: boolean): void;
        didUpdate(): void;
        scheduleUpdateProjection(): void;
        updateProjection: () => void;
        /**
         * Update measurements
         */
        updateSnapshot(): void;
        updateLayout(): void;
        updateScroll(): void;
        resetTransform(): void;
        measure(): Box;
        removeElementScroll(box: Box): Box;
        removeTransform(box: Box): Box;
        /**
         *
         */
        setTargetDelta(delta: Delta): void;
        setOptions(options: ProjectionNodeOptions): void;
        /**
         * Frame calculations
         */
        resolveTargetDelta(): void;
        calcProjection(): void;
        isVisible: boolean;
        hide(): void;
        show(): void;
        scheduleRender(): void;
        /**
         * Animation
         */
        animationProgress: number;
        animationValues?: ResolvedValues | undefined;
        currentAnimation: AnimationPlaybackControls;
        mixTargetDelta: (progress: number) => void;
        setAnimationOrigin(delta: Delta, latestValues: ResolvedValues): void;
        startAnimation(options: AnimationOptions<number>): void;
        applyTransformsToTarget(): void;
        /**
         * Events
         *
         * TODO Replace this with a key-based lookup
         */
        onLayoutWillUpdate(handler: VoidFunction): () => void;
        onLayoutDidUpdate(handler: (data: LayoutUpdateData) => void): () => void;
        /**
         * Shared layout
         */
        sharedNodes: Map<string, NodeStack>;
        registerSharedNode(layoutId: string, node: IProjectionNode): void;
        isLead(): boolean;
        getLead(): IProjectionNode<unknown> | any;
        getStack(): NodeStack | undefined;
        promote(): void;
        getProjectionStyles(): ResolvedValues;
    };
};
export declare function mixAxisDelta(output: AxisDelta, delta: AxisDelta, p: number): void;
